/*
        lzma
                Copyright 2012 by Christian Ghisler (support@ghisler.com)
                This plugin is Freeware

        to be used with Total Commander (www.ghisler.com)
*/

#include <windows.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <io.h>
#include "zlib.h"
#include "resource.h"

char inifilename[MAX_PATH]="wcplugin.ini";  // Unused in this plugin, may be used to save data
BOOL WriteSizeInHeader=false;               // Doesn't work with TC 8.0 or older! -> TC 8.01, version 2.22

/* Error codes returned to calling application */
#define E_END_ARCHIVE     10            /* No more files in archive */
#define E_NO_MEMORY       11            /* Not enough memory */
#define E_BAD_DATA        12            /* Data is bad */
#define E_BAD_ARCHIVE     13            /* CRC error in archive data */
#define E_UNKNOWN_FORMAT  14            /* Archive format unknown */
#define E_EOPEN           15            /* Cannot open existing file */
#define E_ECREATE         16            /* Cannot create file */
#define E_ECLOSE          17            /* Error closing file */
#define E_EREAD           18            /* Error reading from file */
#define E_EWRITE          19            /* Error writing to file */
#define E_SMALL_BUF       20            /* Buffer too small */
#define E_EABORTED        21            /* Function aborted by user */
#define E_NO_FILES        22            /* No files found */
#define E_TOO_MANY_FILES  23            /* Too many files to pack */
#define E_NOT_SUPPORTED   24            /* Function not supported */

/* flags for unpacking */
#define PK_OM_LIST         0
#define PK_OM_EXTRACT      1

/* flags for ProcessFile */
#define PK_SKIP            0            /* Skip this file */
#define PK_TEST            1            /* Test file integrity */
#define PK_EXTRACT         2            /* Extract to disk */

/* Flags passed through ChangeVolProc */
#define PK_VOL_ASK         0            /* Ask user for location of next volume */
#define PK_VOL_NOTIFY      1            /* Notify app that next volume will be unpacked */

/* Flags for packing */

/* For PackFiles */
#define PK_PACK_MOVE_FILES 1    /* Delete original after packing        */
#define PK_PACK_SAVE_PATHS 2    /* Save path names of files             */

/* Returned by GetPackCaps */
#define PK_CAPS_NEW        1    /* Can create new archives              */
#define PK_CAPS_MODIFY     2    /* Can modify exisiting archives        */
#define PK_CAPS_MULTIPLE   4    /* Archive can contain multiple files   */
#define PK_CAPS_DELETE     8    /* Can delete files                     */
#define PK_CAPS_OPTIONS   16    /* Has options dialog                   */
#define PK_CAPS_MEMPACK   32    /* Functions to pack in memory          */
#define PK_CAPS_BY_CONTENT 64   /* Can detect archive by its contents   */
#define PK_CAPS_SEARCHTEXT 128  /* Allow searching in this archive type */

#define BACKGROUND_UNPACK 1      /* Which operations are thread-safe?    */
#define BACKGROUND_PACK 2
#define BACKGROUND_MEMPACK 4

/* Options for memory packing */
#define MEM_OPTIONS_WANTHEADERS 1

/* Returned by PackToMem */
#define MEMPACK_OK         0
#define MEMPACK_DONE       1
#define MEMPACK_ERR       -1

typedef struct {
    char ArcName[260];
    char FileName[260];
    int Flags;
    int PackSize;
    int UnpSize; 
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tHeaderData;

typedef struct {
    char ArcName[1024];
    char FileName[1024];
    int Flags;
    DWORD PackSize;
    int PackSizeHigh;
    DWORD UnpSize; 
    int UnpSizeHigh; 
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
    char Unused[1024];
  } tHeaderDataEx;

typedef struct {
    WCHAR ArcNameW[1024];
    WCHAR FileNameW[1024];
    int Flags;
    DWORD PackSize;
    int PackSizeHigh;
    DWORD UnpSize; 
    int UnpSizeHigh; 
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
    char Unused[1024];
  } tHeaderDataExW;

typedef struct {
    char* ArcName;
    int OpenMode;
    int OpenResult;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tOpenArchiveData;

typedef struct {
    WCHAR* ArcNameW;
    int OpenMode;
    int OpenResult;
    WCHAR* CmtBufW;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tOpenArchiveDataW;

typedef struct {
	int size;
	DWORD PluginInterfaceVersionLow;
	DWORD PluginInterfaceVersionHi;
	char DefaultIniName[MAX_PATH];
} PackDefaultParamStruct;

/* Definition of callback functions called by the DLL
Ask to swap disk for multi-volume archive */
typedef int (__stdcall *tChangeVolProc)(char *ArcName,int Mode);
typedef int (__stdcall *tChangeVolProcW)(WCHAR *ArcName,int Mode);
/* Notify that data is processed - used for progress dialog */
typedef int (__stdcall *tProcessDataProc)(unsigned char *Addr,int Size);
typedef int (__stdcall *tProcessDataProcW)(WCHAR *Addr,int Size);

typedef struct {
  char ArchiveName[260];
  WCHAR ArchiveNameW[260];
  int ArchiveOpen;
  int ArchivePos;
  DWORD ArchiveSize;
  int ArchiveSizeHigh;
  DWORD UnpSize;
  int UnpSizeHigh;
  int ArchiveDate;
  tChangeVolProc ChangeVolProc;
  tChangeVolProcW ChangeVolProcW;
  tProcessDataProc ProcessDataProc;
  tProcessDataProcW ProcessDataProcW;
} ArchiveRec;

int maxarchives=10;

typedef unsigned char Byte;
typedef size_t SizeT;

typedef struct {
  void *(*Alloc)(void *p, SizeT size);
  void (*Free)(void *p, void *address); /* address can be 0 */
} ISzAlloc;

HMODULE dllinst=NULL; 

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
   	if (ul_reason_for_call==DLL_PROCESS_ATTACH) {
		dllinst=(HMODULE)hModule;
	}
	return TRUE;
}

// LZMA 1 functions ==========================================================

#define SZ_OK 0

#define SZ_ERROR_DATA 1
#define SZ_ERROR_MEM 2
#define SZ_ERROR_CRC 3
#define SZ_ERROR_UNSUPPORTED 4
#define SZ_ERROR_PARAM 5
#define SZ_ERROR_INPUT_EOF 6
#define SZ_ERROR_OUTPUT_EOF 7
#define SZ_ERROR_READ 8
#define SZ_ERROR_WRITE 9
#define SZ_ERROR_PROGRESS 10
#define SZ_ERROR_FAIL 11
#define SZ_ERROR_THREAD 12
#define SZ_ERROR_ARCHIVE 16
#define SZ_ERROR_NO_ARCHIVE 17

typedef struct {
	char data[128];
} CLzmaDec;

typedef struct {
	int level;       /*  0 <= level <= 9 */
	DWORD dictSize; /* (1 << 12) <= dictSize <= (1 << 27) for 32-bit version
					  (1 << 12) <= dictSize <= (1 << 30) for 64-bit version
					   default = (1 << 24) */
	int lc;          /* 0 <= lc <= 8, default = 3 */
	int lp;          /* 0 <= lp <= 4, default = 0 */
	int pb;          /* 0 <= pb <= 4, default = 2 */
	int algo;        /* 0 - fast, 1 - normal, default = 1 */
	int fb;          /* 5 <= fb <= 273, default = 32 */
	int btMode;      /* 0 - hashChain Mode, 1 - binTree mode - normal, default = 1 */
	int numHashBytes; /* 2, 3 or 4, default = 4 */
	DWORD mc;        /* 1 <= mc <= (1 << 30), default = 32 */
	unsigned writeEndMark;  /* 0 - do not write EOPM, 1 - write EOPM, default = 0 */
	int numThreads;  /* 1 or 2, default = 2 */
} CLzmaEncProps;

typedef enum
{
  LZMA_FINISH_ANY,   /* finish at any point */
  LZMA_FINISH_END    /* block must be finished at the end */
} ELzmaFinishMode;

typedef enum
{
  LZMA_STATUS_NOT_SPECIFIED,               /* use main error code instead */
  LZMA_STATUS_FINISHED_WITH_MARK,          /* stream was finished with end mark. */
  LZMA_STATUS_NOT_FINISHED,                /* stream was not finished */
  LZMA_STATUS_NEEDS_MORE_INPUT,            /* you must provide more input bytes */
  LZMA_STATUS_MAYBE_FINISHED_WITHOUT_MARK  /* there is probability that stream was finished without end mark */
} ELzmaStatus;

typedef void * CLzmaEncHandle;
typedef int SRes;
typedef unsigned __int64 UInt64;

typedef struct
{
  SRes (*Read)(void *p, void *buf, size_t *size);
    /* if (input(*size) != 0 && output(*size) == 0) means end_of_stream.
       (output(*size) < input(*size)) is allowed */
  FILE *in_file;
} ISeqInStream;

typedef struct
{
  size_t (*Write)(void *p, const void *buf, size_t size);
    /* Returns: result - the number of actually written bytes.
       (result < size) means error */
  FILE *out_file;
} ISeqOutStream;

typedef struct
{
  SRes (*Progress)(void *p, UInt64 inSize, UInt64 outSize);
    /* Returns: result. (result != SZ_OK) means break.
       Value (UInt64)(Int64)-1 for size means unknown value. */
  UInt64 lastSize;
  void* data;
} ICompressProgress;

typedef void (__cdecl *tLzmaDec_Constr)(CLzmaDec* p);
typedef int (__cdecl *tLzmaDec_Allocate)(CLzmaDec* state, const Byte *prop, unsigned int propsSize, ISzAlloc *alloc);
typedef void (__cdecl *tLzmaDec_Init)(CLzmaDec *p);
typedef int (__cdecl *tLzmaDec_DecodeToBuf)(CLzmaDec *p, Byte *dest, SizeT *destLen,
    const Byte *src, SizeT *srcLen, ELzmaFinishMode finishMode, ELzmaStatus *status);
typedef void (__cdecl *tLzmaDec_Free)(CLzmaDec *state, ISzAlloc *alloc);
// compression:
typedef CLzmaEncHandle (__cdecl *tLzmaEnc_Create)(ISzAlloc *alloc);
typedef void (__cdecl *tLzmaEncProps_Init)(CLzmaEncProps *p);
typedef SRes (__cdecl *tLzmaEnc_SetProps)(CLzmaEncHandle pp, const CLzmaEncProps *props2);
typedef SRes (__cdecl *tLzmaEnc_WriteProperties)(CLzmaEncHandle pp, Byte *props, SizeT *size);
typedef SRes (__cdecl *tLzmaEnc_Encode)(CLzmaEncHandle pp, ISeqOutStream *outStream, ISeqInStream *inStream, ICompressProgress *progress,
    ISzAlloc *alloc, ISzAlloc *allocBig);
typedef void (__cdecl *tLzmaEnc_Destroy)(CLzmaEncHandle p, ISzAlloc *alloc, ISzAlloc *allocBig);

tLzmaDec_Constr LzmaDec_Constr=NULL;
tLzmaDec_Allocate LzmaDec_Allocate=NULL;
tLzmaDec_Init LzmaDec_Init=NULL;
tLzmaDec_DecodeToBuf LzmaDec_DecodeToBuf=NULL;
tLzmaDec_Free LzmaDec_Free=NULL;

tLzmaEnc_Create LzmaEnc_Create=NULL;
tLzmaEncProps_Init LzmaEncProps_Init=NULL;
tLzmaEnc_SetProps LzmaEnc_SetProps=NULL;
tLzmaEnc_WriteProperties LzmaEnc_WriteProperties=NULL;
tLzmaEnc_Encode LzmaEnc_Encode=NULL;
tLzmaEnc_Destroy LzmaEnc_Destroy=NULL;

HMODULE lzmalib=NULL;
BOOL loadtried=false;

BOOL LoadLzmaLib()
{
	if (!loadtried) {
		loadtried=true;
		char libname[260];
		GetModuleFileName(dllinst,libname,sizeof(libname)-8);
		char* p=strrchr(libname,'\\');
		if (p)
			p++;
		else
			p=libname;
#if defined _WIN64
		strcpy(p,"tcLZMA64.dll");
#else
		strcpy(p,"tcmdLZMA.dll");
#endif
		int olderrormode=SetErrorMode(0x8001);
		lzmalib=LoadLibrary(libname);
		SetErrorMode(olderrormode);
		if (!lzmalib)
			return false;
		LzmaDec_Constr=(tLzmaDec_Constr)GetProcAddress(lzmalib,"LzmaDec_Constr");
		LzmaDec_Allocate=(tLzmaDec_Allocate)GetProcAddress(lzmalib,"LzmaDec_Allocate");
		LzmaDec_Init=(tLzmaDec_Init)GetProcAddress(lzmalib,"LzmaDec_Init");
		LzmaDec_DecodeToBuf=(tLzmaDec_DecodeToBuf)GetProcAddress(lzmalib,"LzmaDec_DecodeToBuf");
		LzmaDec_Free=(tLzmaDec_Free)GetProcAddress(lzmalib,"LzmaDec_Free");

		LzmaEnc_Create=(tLzmaEnc_Create)GetProcAddress(lzmalib,"LzmaEnc_Create");
		LzmaEncProps_Init=(tLzmaEncProps_Init)GetProcAddress(lzmalib,"LzmaEncProps_Init");
		LzmaEnc_SetProps=(tLzmaEnc_SetProps)GetProcAddress(lzmalib,"LzmaEnc_SetProps");
		LzmaEnc_WriteProperties=(tLzmaEnc_WriteProperties)GetProcAddress(lzmalib,"LzmaEnc_WriteProperties");
		LzmaEnc_Encode=(tLzmaEnc_Encode)GetProcAddress(lzmalib,"LzmaEnc_Encode");
		LzmaEnc_Destroy=(tLzmaEnc_Destroy)GetProcAddress(lzmalib,"LzmaEnc_Destroy");
		
		if (LzmaDec_Constr==NULL || LzmaDec_Allocate==NULL || LzmaDec_Init==NULL ||
			LzmaDec_DecodeToBuf==NULL || LzmaDec_Free==NULL ||
			LzmaEnc_Create==NULL || LzmaEncProps_Init==NULL || LzmaEnc_SetProps==NULL ||
			LzmaEnc_WriteProperties==NULL || LzmaEnc_Encode==NULL || LzmaEnc_Destroy==NULL) {
			FreeLibrary(lzmalib);
			lzmalib=NULL;
		}
	}
	return (lzmalib!=NULL);
}

/***************************************************************/

char* strlcpy(char* p,char*p2,int maxlen)
{
	if ((int)strlen(p2)>=maxlen) {
		strncpy(p,p2,maxlen);
		p[maxlen]=0;
	} else
		strcpy(p,p2);
	return p;
}

WCHAR* wcslcpy(WCHAR* p,WCHAR*p2,int maxlen)
{
	if ((int)wcslen(p2)>=maxlen) {
		wcsncpy(p,p2,maxlen);
		p[maxlen]=0;
	} else
		wcscpy(p,p2);
	return p;
}

CRITICAL_SECTION slotcriticalsection;
BOOL slotcriticalsectioninitialized=false;

#pragma pack(1)

typedef struct {
        char sig[5];
} LZMAHEAD;

/* Keep a list of currently open archives (up to a maximum of maxarchives) */
static ArchiveRec ArchiveList[11];   /* 0 for packing, 1-10 for unpacking! */
                                      /* Important: array must be initialized to all 0! */

int __stdcall OpenArchive(tOpenArchiveData* ArchiveData)
{
    int ArchiveHandle;
    int n;
    int i;
    int ArcHandle;
    int foundslot;
    unsigned short date, time;
    struct stat ftbuf;
    struct tm* fttm;
	LZMAHEAD lzmahead;

    if (!slotcriticalsectioninitialized) {
        slotcriticalsectioninitialized=true;
        InitializeCriticalSection (&slotcriticalsection);
    }
    EnterCriticalSection (&slotcriticalsection);
    foundslot=0;
    for (i=1;i<=maxarchives;i++)
      if (ArchiveList[i].ArchiveOpen==0) {
        foundslot=1;
        ArcHandle=i;
        ArchiveList[ArcHandle].ArchiveOpen=1;
        break;
      }
    LeaveCriticalSection (&slotcriticalsection);

    if (foundslot==0) {
      ArchiveData->OpenResult=E_NO_MEMORY;
      ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    if((ArchiveHandle = _open(ArchiveData->ArcName,_O_BINARY | _O_RDONLY))==-1){
      ArchiveData->OpenResult=E_EOPEN;  
      ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }

    __int64 size64=_lseeki64(ArchiveHandle,0,SEEK_END);
	_lseek(ArchiveHandle,0,SEEK_SET);
	ArchiveList[ArcHandle].ArchiveSize=(DWORD)(size64 & 0xFFFFFFFF);
	ArchiveList[ArcHandle].ArchiveSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);

    n = _read ( ArchiveHandle, &lzmahead, sizeof(lzmahead) );
    if (n != sizeof(lzmahead)) {
      _close(ArchiveHandle);
      ArchiveData->OpenResult=E_EREAD;
	  ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }

	n = _read ( ArchiveHandle, &size64, 8 );
	if (n != 8) {
		_close(ArchiveHandle);
		ArchiveData->OpenResult=E_EREAD;
		ArchiveList[ArcHandle].ArchiveOpen=0;
		return 0;
	}
	ArchiveList[ArcHandle].UnpSize=(DWORD)(size64 & 0xFFFFFFFF);
	ArchiveList[ArcHandle].UnpSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);

    stat(ArchiveData->ArcName,&ftbuf);
    fttm=localtime(&ftbuf.st_mtime);
    time=fttm->tm_hour<<11 | fttm->tm_min<<5 | fttm->tm_sec;
    date=(fttm->tm_year-80)<<9 | (fttm->tm_mon+1)<<5 | fttm->tm_mday;
    ArchiveList[ArcHandle].ArchiveDate=65536*date+time;

    _close(ArchiveHandle);

    strcpy(ArchiveList[ArcHandle].ArchiveName,ArchiveData->ArcName);
    ArchiveList[ArcHandle].ArchiveNameW[0]=0;
    ArchiveList[ArcHandle].ArchiveOpen=1;
    ArchiveList[ArcHandle].ArchivePos=0;
    return ArcHandle;  /* returns pseudohandle */
}

int __stdcall OpenArchiveW(tOpenArchiveDataW* ArchiveData)
{
    int ArchiveHandle;
    int n;
    int i;
    int ArcHandle;
    int foundslot;
    unsigned short date, time;
    struct _stat ftbuf;
    struct tm* fttm;
	LZMAHEAD lzmahead;

    if (!slotcriticalsectioninitialized) {
        slotcriticalsectioninitialized=true;
        InitializeCriticalSection (&slotcriticalsection);
    }
    EnterCriticalSection (&slotcriticalsection);
    foundslot=0;
    for (i=1;i<=maxarchives;i++)
      if (ArchiveList[i].ArchiveOpen==0) {
        foundslot=1;
        ArcHandle=i;
        ArchiveList[ArcHandle].ArchiveOpen=1;
        break;
      }
    LeaveCriticalSection (&slotcriticalsection);

    if (foundslot==0) {
      ArchiveData->OpenResult=E_NO_MEMORY;
      ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    if((ArchiveHandle = _wopen(ArchiveData->ArcNameW,_O_BINARY | _O_RDONLY))==-1){
      ArchiveData->OpenResult=E_EOPEN;  
      ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }

    __int64 size64=_lseeki64(ArchiveHandle,0,SEEK_END);
	_lseek(ArchiveHandle,0,SEEK_SET);
	ArchiveList[ArcHandle].ArchiveSize=(DWORD)(size64 & 0xFFFFFFFF);
	ArchiveList[ArcHandle].ArchiveSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);

    n = _read ( ArchiveHandle, &lzmahead, sizeof(lzmahead) );
    if (n != sizeof(lzmahead)) {
      _close(ArchiveHandle);
      ArchiveData->OpenResult=E_EREAD;
	  ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    
	n = _read ( ArchiveHandle, &size64, 8 );
	if (n != 8) {
		_close(ArchiveHandle);
		ArchiveData->OpenResult=E_EREAD;
		ArchiveList[ArcHandle].ArchiveOpen=0;
		return 0;
	}
    ArchiveList[ArcHandle].UnpSize=(DWORD)(size64 & 0xFFFFFFFF);
    ArchiveList[ArcHandle].UnpSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);

    _wstat(ArchiveData->ArcNameW,&ftbuf);
    fttm=localtime(&ftbuf.st_mtime);
    time=fttm->tm_hour<<11 | fttm->tm_min<<5 | fttm->tm_sec;
    date=(fttm->tm_year-80)<<9 | (fttm->tm_mon+1)<<5 | fttm->tm_mday;
    ArchiveList[ArcHandle].ArchiveDate=65536*date+time;

    _close(ArchiveHandle);

    ArchiveList[ArcHandle].ArchiveName[0]=0;
    wcscpy(ArchiveList[ArcHandle].ArchiveNameW,ArchiveData->ArcNameW);
    ArchiveList[ArcHandle].ArchiveOpen=1;
    ArchiveList[ArcHandle].ArchivePos=0;
    return ArcHandle;  /* returns pseudohandle */
}

int __stdcall CloseArchive(int hArcData)
{
    if (hArcData>0 && hArcData<=maxarchives) {
      ArchiveList[hArcData].ArchiveOpen=0;
      return 0;
    } else  
      return E_ECLOSE;
}

int __stdcall ReadHeader(int hArcData,tHeaderData* HeaderData)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        char* p;
        char* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=strrchr(ArchiveList[hArcData].ArchiveName,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveName; else p++;
        p1=strrchr(p,'.');
        if (p1 != NULL) {
          strncpy(HeaderData->FileName,p,p1-p);
        } else {
          strcpy(HeaderData->FileName,p);  
        }
        strcpy(HeaderData->ArcName,ArchiveList[hArcData].ArchiveName);
        HeaderData->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderData->UnpSize=ArchiveList[hArcData].UnpSize;
        HeaderData->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

int __stdcall ReadHeaderEx(int hArcData,tHeaderDataEx* HeaderDataEx)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        char* p;
        char* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=strrchr(ArchiveList[hArcData].ArchiveName,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveName; else p++;
        p1=strrchr(p,'.');
        if (p1 != NULL) {
          strncpy(HeaderDataEx->FileName,p,p1-p);
        } else {
          strcpy(HeaderDataEx->FileName,p);  
        }
        strcpy(HeaderDataEx->ArcName,ArchiveList[hArcData].ArchiveName);
        HeaderDataEx->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderDataEx->PackSizeHigh=ArchiveList[hArcData].ArchiveSizeHigh;
        HeaderDataEx->UnpSize=ArchiveList[hArcData].UnpSize;
        HeaderDataEx->UnpSizeHigh=ArchiveList[hArcData].UnpSizeHigh;
        HeaderDataEx->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

int __stdcall ReadHeaderExW(int hArcData,tHeaderDataExW* HeaderDataExW)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        WCHAR* p;
        WCHAR* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=wcsrchr(ArchiveList[hArcData].ArchiveNameW,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveNameW; else p++;
        p1=wcsrchr(p,'.');
        if (p1 != NULL) {
          wcsncpy(HeaderDataExW->FileNameW,p,p1-p);
        } else {
          wcscpy(HeaderDataExW->FileNameW,p);  
        }
        wcscpy(HeaderDataExW->ArcNameW,ArchiveList[hArcData].ArchiveNameW);
        HeaderDataExW->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderDataExW->PackSizeHigh=ArchiveList[hArcData].ArchiveSizeHigh;
        HeaderDataExW->UnpSize=ArchiveList[hArcData].UnpSize;
        HeaderDataExW->UnpSizeHigh=ArchiveList[hArcData].UnpSizeHigh;
        HeaderDataExW->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

typedef struct {
     FILE* handle;
} fptr;

typedef fptr* pfptr;

static BOOL myfeof ( FILE* f )
{
   int c = fgetc ( f );
   if (c == EOF) return true;
   ungetc ( c, f );
   return false;
} 

static void *SzAlloc(void *p, SizeT size) { p = p; return malloc(size); }
static void SzFree(void *p, void *address) { p = p; free(address); }
static ISzAlloc g_Alloc = { SzAlloc, SzFree };

#define IN_BUF_SIZE (1 << 16)
#define OUT_BUF_SIZE (1 << 16)

static int DecodeBlockLzma(int Operation,int hArcData,CLzmaDec *state, FILE *outStream, FILE *inStream, __int64 unpackSize)
{
  int aborted=0;
  int thereIsSize = (unpackSize != -1);
  Byte inBuf[IN_BUF_SIZE];
  Byte outBuf[OUT_BUF_SIZE];
  SizeT inPos = 0, inSize = 0, outPos = 0;
  LzmaDec_Init(state);
  for (;;)
  {
	if (inPos == inSize)
	{
	  inSize = fread ( inBuf, 1, IN_BUF_SIZE, inStream);
	  inPos = 0;
	}
	{
	  int res;
	  SizeT inProcessed = inSize - inPos;
	  SizeT outProcessed = OUT_BUF_SIZE - outPos;
	  ELzmaFinishMode finishMode = LZMA_FINISH_ANY;
	  ELzmaStatus status;
	  if (thereIsSize && outProcessed > unpackSize)
	  {
		outProcessed = (SizeT)unpackSize;
		finishMode = LZMA_FINISH_END;
	  }

	  res = LzmaDec_DecodeToBuf(state, outBuf + outPos, &outProcessed,
		inBuf + inPos, &inProcessed, finishMode, &status);
	  inPos += inProcessed;
	  outPos += outProcessed;
	  unpackSize -= outProcessed;

	  if (Operation==PK_EXTRACT || Operation==PK_TEST) {
		if (ArchiveList[hArcData].ProcessDataProc) {
				if (ArchiveList[hArcData].ProcessDataProc(NULL,(int)(outPos))==0)
				aborted=1;
		}
	  }				  

	  if (Operation==PK_EXTRACT) {
		  fwrite(outBuf,1,outPos,outStream);
		  if (ferror(outStream)) {
			return E_EWRITE;
		  } 
	  }

	  outPos = 0;

	  if (res != 0 || thereIsSize && unpackSize == 0) {
		if (res==0)
			return 0;
		else
			return E_EREAD;
	  }

	  if ((inProcessed == 0 && outProcessed == 0) || aborted)
	  {
        if (aborted)
          return E_EABORTED;
		if (thereIsSize || status != LZMA_STATUS_FINISHED_WITH_MARK)
		  return E_EREAD;
		if (res==0)
		  return 0;
		else
		  return E_EREAD;
	  }
	}
  }
}

int __stdcall ProcessFileAW(BOOL unicode,int hArcData,int Operation,
						   char* DestPathA,char* DestNameA,
						   WCHAR* DestPathW,WCHAR* DestNameW)
{
    if (hArcData<1 || hArcData>maxarchives)
        return E_EREAD;

    if (Operation==PK_SKIP)
        return 0;
        
    if (ArchiveList[hArcData].ArchiveOpen && ArchiveList[hArcData].ArchiveNameW[0] &&
        ((Operation==PK_EXTRACT && (DestNameA!=NULL || DestNameW !=NULL)) || (Operation==PK_TEST))) {

		if (!LoadLzmaLib())
			return E_EREAD;

        CLzmaDec state;
        FILE *fp_w=NULL;
		FILE *inStream;
        int aborted=0;
		int res;
		BOOL smallMode = false;
		int verbosity = 0;
		int streamNo = 0;

		if (unicode) {
			if (ArchiveList[hArcData].ProcessDataProcW) {
				if (ArchiveList[hArcData].ProcessDataProcW(DestNameW ? DestNameW:ArchiveList[hArcData].ArchiveNameW,0)==0)
					return E_EABORTED;
			}
		} else {
			if (ArchiveList[hArcData].ProcessDataProc) {
				if (ArchiveList[hArcData].ProcessDataProc((unsigned char*)(DestNameA ? DestNameA:ArchiveList[hArcData].ArchiveName),0)==0)
					return E_EABORTED;
			}
		}
        
		if (Operation==PK_EXTRACT)
			if (unicode) {
                if ((fp_w = _wfopen(DestNameW,L"wb"))==NULL)
                        return E_ECREATE;
			} else {
                if ((fp_w = fopen(DestNameA,"wb"))==NULL)
                        return E_ECREATE;
			}

		if (unicode)
			inStream = _wfopen ( ArchiveList[hArcData].ArchiveNameW, L"rb" );
		else
			inStream = fopen ( ArchiveList[hArcData].ArchiveName, "rb" );
		if ( inStream == NULL ) {
			if (Operation==PK_EXTRACT)
				fclose(fp_w);
			return E_EOPEN;
		}

		char head[5+8];
		memset(&state,0,sizeof(state));
		LzmaDec_Constr(&state);
		__int64 unpackSize;
		int n = fread(head, 1, 5, inStream);
		int m = fread(&unpackSize, 1, 8, inStream);

		if (n != 5 || m!=8) {
		  res=E_EREAD;
		} else {
			res=LzmaDec_Allocate(&state,(const unsigned char*)&head,5,&g_Alloc);
			if (res)
				res=E_NO_MEMORY;
			else {
				res=DecodeBlockLzma(Operation,hArcData,&state, fp_w, inStream, unpackSize);
				LzmaDec_Free(&state,&g_Alloc);
			}
		}

		fclose (inStream); 

        if (Operation==PK_EXTRACT)
                fclose(fp_w);
        if (aborted)
                return E_EABORTED;
        else if (res!=0) {
			if(unicode){
				if (DestNameW)
					DeleteFileW(DestNameW);
			} else {
				if (DestNameA)
					DeleteFile(DestNameA);
			}
                return res;
        } else
                return 0;
    } else
        return E_EREAD;
}

int __stdcall ProcessFile(int hArcData,int Operation,char* DestPath,char* DestName)
{
	return ProcessFileAW(false,hArcData,Operation,DestPath,DestName,NULL,NULL);
}

int __stdcall ProcessFileW(int hArcData,int Operation,WCHAR* DestPath,WCHAR* DestName)
{
	return ProcessFileAW(true,hArcData,Operation,NULL,NULL,DestPath,DestName);
}

void __stdcall SetChangeVolProc(int hArcData,tChangeVolProc pChangeVolProc)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ChangeVolProc=pChangeVolProc;
}

void __stdcall SetProcessDataProc(int hArcData,tProcessDataProc pProcessDataProc)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ProcessDataProc=pProcessDataProc;    
}

void __stdcall SetChangeVolProcW(int hArcData,tChangeVolProcW pChangeVolProcW)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ChangeVolProcW=pChangeVolProcW;
}

void __stdcall SetProcessDataProcW(int hArcData,tProcessDataProcW pProcessDataProcW)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ProcessDataProcW=pProcessDataProcW;
}

int FileExists(char *FileName)
{
    FILE *fp_r;
    int res;
    fp_r = fopen(FileName,"rb");
    res=fp_r!=NULL;
    if (res) fclose(fp_r);
    return res;
}

int FileExistsW(WCHAR *FileName)
{
    FILE *fp_r;
    int res;
    fp_r = _wfopen(FileName,L"rb");
    res=fp_r!=NULL;
    if (res) fclose(fp_r);
    return res;
}

int RenToTemp(char *FileName,char *RenamedName)    /* returns RenamedName */
{
    char *p;
    char ch1;
    char ch2;
    strcpy(RenamedName,FileName);
    p=RenamedName;
    while (p[0]) p++;
    strncat(RenamedName,".#00",259);
    p+=2;
    for (ch1='0';ch1<='9';ch1++) {
      p[0]=ch1;
      for (ch2='0';ch2<='9';ch2++) {
        p[1]=ch2;
        if (!rename(FileName,RenamedName))
          return 1;
      }
    }
    return 0;    /* rename failed! */
}

int RenToTempW(WCHAR *FileName,WCHAR *RenamedName)    /* returns RenamedName */
{
    WCHAR *p;
    WCHAR ch1;
    WCHAR ch2;
    wcscpy(RenamedName,FileName);
    p=RenamedName;
    while (p[0]) p++;
    wcsncat(RenamedName,L".#00",259);
    p+=2;
    for (ch1='0';ch1<='9';ch1++) {
      p[0]=ch1;
      for (ch2='0';ch2<='9';ch2++) {
        p[1]=ch2;
        if (MoveFileW(FileName,RenamedName))
          return 1;
      }
    }
    return 0;    /* rename failed! */
}

SRes ReadFunction(void *p, void *buf, size_t *size) {
    /* if (input(*size) != 0 && output(*size) == 0) means end_of_stream.
       (output(*size) < input(*size)) is allowed */
	ISeqInStream* instream=(ISeqInStream*)p;
	*size = fread (buf,1,*size,instream->in_file);
	if (ferror (instream->in_file)) {
		return SZ_ERROR_READ;
	}
	return SZ_OK;
}

size_t WriteFunction(void *p, const void *buf, size_t size) {
    /* Returns: result - the number of actually written bytes.
       (result < size) means error */
	ISeqOutStream* outstream=(ISeqOutStream*)p;
	
	fwrite (buf, 1, size, outstream->out_file);
	if (ferror (outstream->out_file)) {
		return 0;
	}
	return size;
}

SRes ProgressFunction(void *p, UInt64 inSize, UInt64 outSize) {
    /* Returns: result. (result != SZ_OK) means break.
       Value (UInt64)(Int64)-1 for size means unknown value. */
	ICompressProgress* progress=(ICompressProgress*)p;

	if (ArchiveList[0].ProcessDataProc) {
		if (ArchiveList[0].ProcessDataProc(NULL,(int)(inSize-progress->lastSize))==0) {
			return SZ_ERROR_PROGRESS;
		}
		progress->lastSize=inSize;
	}
	return NULL;	  
}

#define LZMA_PROPS_SIZE 5

int lzma_compress (FILE *in_file, FILE *out_file, int level, DWORD dictsize)
{
	/* initialize xz encoder */
	CLzmaEncHandle enc;
	CLzmaEncProps props;

	enc = LzmaEnc_Create(&g_Alloc);
	if (enc == 0)
		return E_NO_MEMORY;
	
	LzmaEncProps_Init(&props);
	props.level=level;
	props.dictSize=dictsize;
	props.writeEndMark=true;
	int ret = LzmaEnc_SetProps(enc, &props);
	if (ret!=0) {
		LzmaEnc_Destroy(enc, &g_Alloc, &g_Alloc);
		return E_BAD_DATA;
	}

    Byte header[LZMA_PROPS_SIZE + 8];
    size_t headerSize = LZMA_PROPS_SIZE;
    UInt64 fileSize;
    int i;

    int res = LzmaEnc_WriteProperties(enc, header, &headerSize);

    fileSize=_lseeki64(in_file->_file,0,SEEK_END);
	_lseek(in_file->_file,0,SEEK_SET);

    for (i = 0; i < 8; i++)
      header[headerSize++] = (Byte)(fileSize >> (8 * i));
	fwrite (header, 1, headerSize, out_file);
	if (ferror (out_file)) {
		LzmaEnc_Destroy(enc, &g_Alloc, &g_Alloc);
		ret = E_EWRITE;
	}

	ISeqOutStream out_stream;
	out_stream.out_file=out_file;
	out_stream.Write=WriteFunction;

	ISeqInStream in_stream;
	in_stream.in_file=in_file;
	in_stream.Read=ReadFunction;

	ICompressProgress progress;
	progress.Progress=ProgressFunction;
	progress.lastSize=0;

	res = LzmaEnc_Encode(enc, &out_stream, &in_stream,
		&progress, &g_Alloc, &g_Alloc);

	LzmaEnc_Destroy(enc, &g_Alloc, &g_Alloc);

	switch (res) {
	case SZ_OK:
		return 0;
	case SZ_ERROR_READ:
	case SZ_ERROR_INPUT_EOF:
		return E_EREAD;
	case SZ_ERROR_WRITE:
	case SZ_ERROR_OUTPUT_EOF:
		return E_EWRITE;
	case SZ_ERROR_MEM:
	case SZ_ERROR_THREAD:
		return E_NO_MEMORY;
	case SZ_ERROR_DATA:
		return E_BAD_DATA;
	case SZ_ERROR_UNSUPPORTED:
	case SZ_ERROR_PARAM:
		return E_NOT_SUPPORTED;
	case SZ_ERROR_PROGRESS:
		return E_EABORTED;
	}
	return E_BAD_DATA;
} 

int __stdcall PackFiles(char *PackedFile,char *SubPath,char *SrcPath,
      char *AddList,int Flags)
{
	char FullName[MAX_PATH];
	char OldName[MAX_PATH];
	FILE *fp_r,*fp_w;
	char *p;
	int retval=0;

	p=AddList;
	if (!p || !p[0]) return E_NO_FILES;
	while (p[0]) p++;
	p++;
	if (p[0]) return E_TOO_MANY_FILES;

	if (!LoadLzmaLib())
		return E_NOT_SUPPORTED;

	// Start to pack 
	int level = GetPrivateProfileInt("lzma","CompressionRate",6,inifilename);
	if (level<=0) level=1;
	if (level>=9) level=9;
	DWORD dictsize = GetPrivateProfileInt("lzma","DictSize",16*1024*1024,inifilename);

	strlcpy(FullName,SrcPath,MAX_PATH);
	strncat(FullName,AddList,MAX_PATH);

	BOOL filethere=FileExists(PackedFile);
	if (filethere) {
		// Make sure the added filename is the same as the archive name without extension!
		char *p;
		char TestName[260];
		strcpy(TestName,PackedFile);
		p=strrchr(TestName,'.');
		if (p) p[0]='\0';
		p=strrchr(TestName,'\\');
		if (p) p++;
		if (!p || stricmp(p,AddList)!=0)
				return E_TOO_MANY_FILES;
		filethere=RenToTemp(PackedFile,OldName);
	}        

	if((fp_r = fopen(FullName,"rb"))==NULL) {
		if (filethere)
			rename(OldName,PackedFile);
		return E_EOPEN;
	}        

	if ((fp_w = fopen(PackedFile,"wb"))==NULL) {
		fclose(fp_r);
		if (filethere)
			rename(OldName,PackedFile);
		return E_ECREATE;
    }
	
	retval=lzma_compress(fp_r,fp_w,level,dictsize);

	fclose(fp_w);
	fclose(fp_r);
	// Error occurred -> delete new archive and restore old (if exists)
	if (retval) {
		remove(PackedFile);
		if (filethere)
			rename(OldName,PackedFile);
	} else {
		if (filethere)
			unlink(OldName);
		// Move to archive - only if packing was successful!
		if (Flags & PK_PACK_MOVE_FILES)
			unlink(FullName);
	}
	return retval; 
}

int __stdcall PackFilesW(WCHAR *PackedFile,WCHAR *SubPath,WCHAR *SrcPath,
      WCHAR *AddList,int Flags)
{
	WCHAR FullName[MAX_PATH];
	WCHAR OldName[MAX_PATH];
	FILE *fp_r,*fp_w;
	WCHAR *p;
	int retval=0;

	p=AddList;
	if (!p || !p[0]) return E_NO_FILES;
	while (p[0]) p++;
	p++;
	if (p[0]) return E_TOO_MANY_FILES;

	if (!LoadLzmaLib())
		return E_EREAD;

	// Start to pack 
	int level = GetPrivateProfileInt("lzma","CompressionRate",6,inifilename);
	if (level<=0) level=1;
	if (level>=9) level=9;
	DWORD dictsize = GetPrivateProfileInt("lzma","DictSize",16*1024*1024,inifilename);

	wcslcpy(FullName,SrcPath,MAX_PATH);
	wcsncat(FullName,AddList,MAX_PATH);

	BOOL filethere=FileExistsW(PackedFile);
	if (filethere) {
		// Make sure the added filename is the same as the archive name without extension!
		WCHAR *p;
		WCHAR TestName[260];
		wcscpy(TestName,PackedFile);
		p=wcsrchr(TestName,'.');
		if (p) p[0]='\0';
		p=wcsrchr(TestName,'\\');
		if (p) p++;
		if (!p || wcsicmp(p,AddList)!=0)
				return E_TOO_MANY_FILES;
		filethere=RenToTempW(PackedFile,OldName);
	}        

	if((fp_r = _wfopen(FullName,L"rb"))==NULL) {
		if (filethere)
			MoveFileW(OldName,PackedFile);
		return E_EOPEN;
	}        

	if ((fp_w = _wfopen(PackedFile,L"wb"))==NULL) {
		fclose(fp_r);
		if (filethere)
			MoveFileW(OldName,PackedFile);
		return E_ECREATE;
    }
	
	retval=lzma_compress(fp_r,fp_w,level,dictsize);

	fclose(fp_w);
	fclose(fp_r);
	// Error occurred -> delete new archive and restore old (if exists)
	if (retval) {
		DeleteFileW(PackedFile);
		if (filethere)
			MoveFileW(OldName,PackedFile);
	} else {
		if (filethere)
			DeleteFileW(OldName);
		// Move to archive - only if packing was successful!
		if (Flags & PK_PACK_MOVE_FILES)
			DeleteFileW(FullName);
	}
	return retval; 
}

int __stdcall DeleteFiles(char *PackedFile,char *DeleteList)
{
  // Since this archive contains only one file anyway, just delete the whole archive!
  if (DeleteList && DeleteList[0])
    if (unlink(PackedFile))
      return E_EWRITE;
    else
      return 0;
  return E_BAD_DATA;
}

int __stdcall DeleteFilesW(WCHAR *PackedFile,WCHAR *DeleteList)
{
  // Since this archive contains only one file anyway, just delete the whole archive!
  if (DeleteList && DeleteList[0])
    if (!DeleteFileW(PackedFile))
      return E_EWRITE;
    else
      return 0;
  return E_BAD_DATA;
}

int __stdcall GetPackerCaps()
{
  return PK_CAPS_NEW | PK_CAPS_MODIFY | PK_CAPS_DELETE | PK_CAPS_OPTIONS | PK_CAPS_MEMPACK | PK_CAPS_SEARCHTEXT;
}

HANDLE hBrush;

#define COUNTOF(I) sizeof(I)/sizeof(I[0])

int sizes[]={64,128,256,512,1024,2*1024,3*1024,4*1024,
	6*1024,8*1024,10*1024,12*1024,16*1024,20*1024,24*1024,
	28*1024,32*1024,40*1024,48*1024,64*1024,72*1024,96*1024,128*1024};

BOOL CALLBACK DialogProc(HWND hwndDlg,UINT  uMsg,WPARAM wParam,LPARAM  lParam)
{
	int i,dictidx,dictsize;
	char num[32];
	char ch;
	num[1]='\0';
	switch (uMsg) {
	case WM_INITDIALOG:
		for (ch='1';ch<='9';ch++) {
		   num[0]=ch;
		   SendDlgItemMessage(hwndDlg,ID_RATE,CB_ADDSTRING,0,(LPARAM)&num);
		}
		i=GetPrivateProfileInt("lzma","CompressionRate",6,inifilename);
		SendDlgItemMessage(hwndDlg,ID_RATE,CB_SETCURSEL,i-1,0);

		dictsize = GetPrivateProfileInt("lzma","DictSize",16*1024*1024,inifilename);
		dictidx = 0;
		for (i=0;i<COUNTOF(sizes);i++) {
		   int j=sizes[i];
		   if (j*1024<=dictsize)
			   dictidx=i;
		   if (j<1024) {
			   itoa(j,num,10);
			   strcat(num, " kB");
		   } else {
			   itoa(j/1024,num,10);
			   strcat(num, " MB");
		   }
		   SendDlgItemMessage(hwndDlg,ID_DICT,CB_INSERTSTRING,i,(LPARAM)&num);
		}
		SendDlgItemMessage(hwndDlg,ID_DICT,CB_SETCURSEL,dictidx,0);

		hBrush=CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
		return TRUE;
		break;
     case WM_CTLCOLORDLG:
     case WM_CTLCOLORSTATIC:
     case WM_CTLCOLORBTN:
		SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
		SetTextColor((HDC)wParam,GetSysColor(COLOR_BTNTEXT));
		return (BOOL)hBrush;
		break;
     case WM_COMMAND:
		switch (wParam) {
		case ID_OK:    /* OK  */
			num[0]=(char)('1'+SendDlgItemMessage(hwndDlg,ID_RATE,CB_GETCURSEL,0,0));
			WritePrivateProfileString("lzma","CompressionRate",num,inifilename);
			dictidx=SendDlgItemMessage(hwndDlg,ID_DICT,CB_GETCURSEL,0,0);
			if (dictidx>=0 && dictidx<COUNTOF(sizes)) {
				dictsize=sizes[dictidx]*1024;
				itoa(dictsize,num,10);
				WritePrivateProfileString("lzma","DictSize",num,inifilename);
			} else {
				WritePrivateProfileString("lzma","DictSize",NULL,inifilename);
			}
			DeleteObject(hBrush);
			EndDialog(hwndDlg,1);
         case 2:    /* Cancel */
           DeleteObject(hBrush);
           EndDialog(hwndDlg,0);
       }
     break;
   }
   return FALSE;
}

void __stdcall ConfigurePacker(HWND ParentHandle, HINSTANCE hinstance)
{	
   /* You may ask for packing (NOT unpacking!) parameters here */
   DialogBox(hinstance,"CONFDLG",ParentHandle,(DLGPROC)&DialogProc);
}

void __stdcall PackSetDefaultParams(PackDefaultParamStruct* dps)
{
	strlcpy(inifilename,dps->DefaultIniName,MAX_PATH-1);
	WriteSizeInHeader=(dps->PluginInterfaceVersionHi>2 ||
		(dps->PluginInterfaceVersionHi>=2 && dps->PluginInterfaceVersionLow>=22));
}

#define INDATASIZE 65536
#define OUTDATASIZE 65536

typedef struct {
	char inbuf[INDATASIZE];
	char outbuf[OUTDATASIZE];
	__int64 TotalDataRead;
	__int64 TotalDataWritten;
	size_t indatapos;
	size_t outdatapos;
	CRITICAL_SECTION DataSection;
	HANDLE hDataAvailableEvent;
	HANDLE hPackContinueEvent;
	int error;
	BOOL ineof;
	BOOL outHasMoreData;
	BOOL aborted;
} threaddata,*threaddataptr;

size_t mins(size_t a,size_t b) {
	if (a<b)
		return a;
	else
		return b;
}

SRes ReadMemFunction(void *p, void *buf, size_t *size) {
    /* if (input(*size) != 0 && output(*size) == 0) means end_of_stream.
       (output(*size) < input(*size)) is allowed */
	ISeqInStream* instream=(ISeqInStream*)p;
	threaddataptr tdp=(threaddataptr)instream->in_file;

	EnterCriticalSection(&tdp->DataSection);
	BOOL needwait=(tdp->indatapos==0 && !tdp->ineof);
	if (needwait) {
		ResetEvent(tdp->hDataAvailableEvent);   // we must reset it WITHIN the critical section, because there is no data!
		LeaveCriticalSection(&tdp->DataSection);
		WaitForSingleObject(tdp->hDataAvailableEvent,INFINITE);
		EnterCriticalSection(&tdp->DataSection);
	}
	if (tdp->aborted) {
		*size=0;
		return 0;
	}
	size_t copylen=mins(*size,tdp->indatapos);
	if (copylen>0) {
		memcpy(buf,tdp->inbuf,copylen);
		*size=copylen;
		tdp->TotalDataRead+=copylen;
		size_t remain=tdp->indatapos-copylen;
		if (remain>0)
			memmove(tdp->inbuf,tdp->inbuf+copylen,remain);
		tdp->indatapos=remain;
	} else
		*size=0;
	SetEvent(tdp->hPackContinueEvent);
	LeaveCriticalSection(&tdp->DataSection);
	return SZ_OK;
}

size_t WriteMemFunction(void *p, const void *buf, size_t size) {
    /* Returns: result - the number of actually written bytes.
       (result < size) means error */
	ISeqOutStream* outstream=(ISeqOutStream*)p;
	threaddataptr tdp=(threaddataptr)outstream->out_file;
	
	EnterCriticalSection(&tdp->DataSection);
	size_t bufpos=0;
	do {
		size_t copylen=min(size,OUTDATASIZE-tdp->outdatapos);
		if (copylen>0) {
			memcpy(tdp->outbuf+tdp->outdatapos,(char*)buf+bufpos,copylen);
			bufpos+=copylen;
			size-=copylen;
			tdp->outdatapos+=copylen;
			tdp->TotalDataWritten+=copylen;
			SetEvent(tdp->hPackContinueEvent);
		}
		BOOL needwait=(size>0);
		if (needwait) {
			tdp->outHasMoreData=true;
			LeaveCriticalSection(&tdp->DataSection);
			WaitForSingleObject(tdp->hDataAvailableEvent,INFINITE);
			EnterCriticalSection(&tdp->DataSection);
		}
	} while (size>0 && tdp->error<=1);
	tdp->outHasMoreData=false;
	LeaveCriticalSection(&tdp->DataSection);
	return bufpos;
}

SRes MemProgressFunction(void *p, UInt64 inSize, UInt64 outSize) {
    /* Returns: result. (result != SZ_OK) means break.
       Value (UInt64)(Int64)-1 for size means unknown value. */
	ICompressProgress* progress=(ICompressProgress*)p;
	threaddataptr tdp=(threaddataptr)progress->data;

	if (tdp->aborted)
		return SZ_ERROR_PROGRESS;
	return 0;
}

DWORD WINAPI PackThreadProc(LPVOID lpParameter)
{
	threaddataptr tdp=(threaddataptr)lpParameter;
	int waitret=WaitForSingleObject(tdp->hDataAvailableEvent,INFINITE);
	if (waitret==WAIT_OBJECT_0) {
		// Start to pack 
		int level = GetPrivateProfileInt("lzma","CompressionRate",6,inifilename);
		if (level<=0) level=1;
		if (level>=9) level=9;
		DWORD dictsize = GetPrivateProfileInt("lzma","DictSize",16*1024*1024,inifilename);
		CLzmaEncHandle enc;
		CLzmaEncProps props;

		enc = LzmaEnc_Create(&g_Alloc);
		if (enc == 0) {
			tdp->error=E_NO_MEMORY;
			return tdp->error;
		}
		LzmaEncProps_Init(&props);
		props.level=level;
		props.dictSize=dictsize;
		props.writeEndMark=true;
		int ret = LzmaEnc_SetProps(enc, &props);
		if (ret!=0) {
			LzmaEnc_Destroy(enc, &g_Alloc, &g_Alloc);
			tdp->error=E_BAD_DATA;
			return tdp->error;
		}

		Byte header[LZMA_PROPS_SIZE + 8];
		size_t headerSize = LZMA_PROPS_SIZE;
		int i;
		int res = LzmaEnc_WriteProperties(enc, header, &headerSize);

		for (i = 0; i < 8; i++)
		  header[headerSize++] = 0xFF;
		memcpy(tdp->outbuf,header,headerSize);
		tdp->outdatapos=headerSize;
		tdp->TotalDataWritten=headerSize;

		ISeqOutStream out_stream;
		out_stream.out_file=(FILE*)tdp;
		out_stream.Write=WriteMemFunction;

		ISeqInStream in_stream;
		in_stream.in_file=(FILE*)tdp;
		in_stream.Read=ReadMemFunction;

		ICompressProgress progress;
		progress.Progress=MemProgressFunction;
		progress.lastSize=0;
		progress.data=tdp;

		res = LzmaEnc_Encode(enc, &out_stream, &in_stream,
			&progress, &g_Alloc, &g_Alloc);

		LzmaEnc_Destroy(enc, &g_Alloc, &g_Alloc);

		switch (res) {
		case SZ_OK:
			tdp->error=1; break;  // signals successful end!
		case SZ_ERROR_READ:
		case SZ_ERROR_INPUT_EOF:
			tdp->error=E_EREAD; break;
		case SZ_ERROR_WRITE:
		case SZ_ERROR_OUTPUT_EOF:
			tdp->error=E_EWRITE; break;
		case SZ_ERROR_MEM:
		case SZ_ERROR_THREAD:
			tdp->error=E_NO_MEMORY; break;
		case SZ_ERROR_DATA:
			tdp->error=E_BAD_DATA; break;
		case SZ_ERROR_UNSUPPORTED:
		case SZ_ERROR_PARAM:
			tdp->error=E_NOT_SUPPORTED; break;
		case SZ_ERROR_PROGRESS:
			tdp->error=E_EABORTED; break;
		default:
			tdp->error=E_BAD_DATA;
		}
	}
	return tdp->error;
}

HANDLE __stdcall StartMemPack(int options,char* filename) {
	
	if (!LoadLzmaLib())
		return NULL;
	DWORD pThreadId;
	threaddataptr tdp=(threaddataptr)malloc(sizeof(threaddata));
	memset(tdp,0,sizeof(threaddata));
	tdp->hDataAvailableEvent=CreateEvent(NULL,false,false,NULL);
	tdp->hPackContinueEvent=CreateEvent(NULL,false,false,NULL);
	InitializeCriticalSection(&tdp->DataSection);
	HANDLE hAbortThread=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)&PackThreadProc,
									tdp,0,&pThreadId);
	if (hAbortThread) {
		CloseHandle(hAbortThread);
		return(HANDLE)tdp;
	} else {
		free(tdp);
		return NULL;
	}
}

int __stdcall PackToMem(HANDLE hMemPack,char* BufIn,int InLen,int* Taken,
      char* BufOut,int OutLen,int* Written,int* SeekBy)
{
	threaddataptr tdp=(threaddataptr)hMemPack;
	EnterCriticalSection(&tdp->DataSection);
	// Check whether we must wait:
	// If there is in data but no space in inbuf,
	// and outdata is also empty, and pack still active
	if (InLen==0)
		tdp->ineof=true;
	BOOL mustwait=(tdp->error<=1 &&
		((InLen>0 && tdp->indatapos==INDATASIZE) || (InLen==0 && tdp->outdatapos==0 && !tdp->outHasMoreData)));
	if (mustwait) {
		LeaveCriticalSection(&tdp->DataSection);
		WaitForSingleObject(tdp->hPackContinueEvent,10);
		EnterCriticalSection(&tdp->DataSection);
	}
	size_t copylen1=0;
	if (InLen>0 && tdp->indatapos<INDATASIZE) {
		// copy in data to inbuf
		copylen1=mins(InLen,INDATASIZE-tdp->indatapos);
		if (copylen1>0) {
			memcpy(tdp->inbuf+tdp->indatapos,BufIn,copylen1);
			tdp->indatapos+=copylen1;
		} else
			copylen1=0;
	}
	*Taken=copylen1;
	size_t copylen2=0;
	if (OutLen>0 && tdp->outdatapos>0) {
		// copy out data from outbuf
		copylen2=mins(OutLen,tdp->outdatapos);
		if (copylen2>0) {
			memcpy(BufOut,tdp->outbuf,copylen2);
			int remain=tdp->outdatapos-copylen2;
			if (remain>0) {
				// move remaining data to start!
				memmove(tdp->outbuf,tdp->outbuf+copylen2,remain);
				tdp->outdatapos=remain;
			} else
				tdp->outdatapos=0;
		}
	}
	*Written=copylen2;
	int ret;
	if (tdp->error==1) {  // end reached!
		if (tdp->outHasMoreData || tdp->outdatapos>0)
			ret=0;   // more data to come!
		else {
			if (copylen2==0) {
				ret=1;       // final block!
				// write the size!
				if (WriteSizeInHeader &&
				   tdp->TotalDataWritten<0xFFFFFFFF) {
					*SeekBy=(int)(-tdp->TotalDataWritten+5);
					UInt64 fileSize=tdp->TotalDataRead;
					for (int i = 0; i < 8; i++)
						BufOut[i] = (Byte)(fileSize >> (8 * i));
					*Written=8;
				}
			} else
				ret=0;
		}
	} else
		ret=tdp->error;
	
	if (copylen1 || copylen2 || InLen==0)
		SetEvent(tdp->hDataAvailableEvent);   // within the critical section!
	LeaveCriticalSection(&tdp->DataSection);
	return ret;
}

int __stdcall DoneMemPack(HANDLE hMemPack) {
	threaddataptr tdp=(threaddataptr)hMemPack;
	EnterCriticalSection(&tdp->DataSection);
	if (tdp->error==0) {
		tdp->aborted=true;
		do {
			// simulate continued data
			tdp->indatapos=INDATASIZE;
			tdp->outdatapos=0;
			SetEvent(tdp->hDataAvailableEvent);   // within the critical section!
			LeaveCriticalSection(&tdp->DataSection);
			Sleep(50);
			EnterCriticalSection(&tdp->DataSection);
		} while (tdp->error==0);

	}
	LeaveCriticalSection(&tdp->DataSection);
	DeleteCriticalSection(&tdp->DataSection);
	CloseHandle(tdp->hDataAvailableEvent);
	CloseHandle(tdp->hPackContinueEvent);
	free(tdp);
	return 0;
}


